@@ -23,12 +23,14 @@ different handlers to different XMPP stanzas and namespaces
 from __future__ import unicode_literals
 from . import simplexml
 import sys
+import time
 import locale
 import re
 import uuid
+import copy
 from xml.parsers.expat import ExpatError
 from .plugin import PlugIn
-from .protocol import (NS_STREAMS, NS_XMPP_STREAMS, NS_HTTP_BIND, Iq, Presence,
+from .protocol import (NS_DELAY2, NS_STREAMS, NS_XMPP_STREAMS, NS_HTTP_BIND, Iq, Presence,
         Message, Protocol, Node, Error, ERR_FEATURE_NOT_IMPLEMENTED, StreamError)
 import logging
 
@@ -387,7 +389,7 @@ class XMPPDispatcher(PlugIn):
 
     def UnregisterCycleHandler(self, handler):
         """
-        Unregister handler that will is called on every Dispatcher.Process() call
+        Unregister handler that will be called on every Dispatcher.Process() call
         """
         if handler in self._cycleHandlers:
             self._cycleHandlers.remove(handler)
@@ -569,14 +571,25 @@ class XMPPDispatcher(PlugIn):
                 if self._owner._registered_name and not stanza.getAttr('from'):
                     stanza.setAttr('from', self._owner._registered_name)
 
+        self._owner.Connection.send(stanza, now)
+
         # If no ID then it is a whitespace
         if self.sm and self.sm.enabled and ID:
+            # add timestamp to message stanza in queue
+            if (stanza.getName() == 'message' and
+                    stanza.getType() in ('chat', 'groupchat')):
+                timestamp = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
+                attrs = {'stamp': timestamp}
+                if stanza.getType() != 'groupchat':
+                    # Dont leak our JID to Groupchats
+                    attrs['from'] = stanza.getAttr('from')
+                stanza.addChild('delay', namespace=NS_DELAY2, attrs=attrs)
             self.sm.uqueue.append(stanza)
             self.sm.out_h += 1
+
             if len(self.sm.uqueue) > self.sm.max_queue:
                 self.sm.request_ack()
 
-        self._owner.Connection.send(stanza, now)
         return ID
 
 
