@@ -24,6 +24,7 @@ from .simplexml import Node, NodeBuilder
 import time
 import string
 import hashlib
+from base64 import b64encode
 
 def ascii_upper(s):
     return s.upper()
@@ -52,6 +53,7 @@ NS_BYTESTREAM     = 'http://jabber.org/protocol/bytestreams'          # XEP-0065
 NS_CAPS           = 'http://jabber.org/protocol/caps'                 # XEP-0115
 NS_CAPTCHA        = 'urn:xmpp:captcha'                                # XEP-0158
 NS_CARBONS        = 'urn:xmpp:carbons:2'                              # XEP-0280
+NS_CHATMARKERS    = 'urn:xmpp:chat-markers:0'                         # XEP-0333
 NS_CHATSTATES     = 'http://jabber.org/protocol/chatstates'           # XEP-0085
 NS_CHATTING       = 'http://jabber.org/protocol/chatting'             # XEP-0194
 NS_CLIENT         = 'jabber:client'
@@ -70,6 +72,7 @@ NS_DIALBACK       = 'jabber:server:dialback'
 NS_DISCO          = 'http://jabber.org/protocol/disco'
 NS_DISCO_INFO     = NS_DISCO + '#info'
 NS_DISCO_ITEMS    = NS_DISCO + '#items'
+NS_EME            = 'urn:xmpp:eme:0'                                  # XEP-0380
 NS_ENCRYPTED      = 'jabber:x:encrypted'                              # XEP-0027
 NS_ESESSION       = 'http://www.xmpp.org/extensions/xep-0116.html#ns'
 NS_ESESSION_INIT  = 'http://www.xmpp.org/extensions/xep-0116.html#ns-init' # XEP-0116
@@ -84,7 +87,9 @@ NS_GROUPCHAT      = 'gc-1.0'
 NS_MSG_HINTS      = 'urn:xmpp:hints'                                  # XEP-0280
 NS_HTTP_AUTH      = 'http://jabber.org/protocol/http-auth'            # XEP-0070
 NS_HTTP_BIND      = 'http://jabber.org/protocol/httpbind'             # XEP-0124
+NS_HTTPUPLOAD     = 'urn:xmpp:http:upload'                            # XEP-0363
 NS_IBB            = 'http://jabber.org/protocol/ibb'
+NS_IDLE           = 'urn:xmpp:idle:1'                                 # XEP-0319
 NS_INVISIBLE      = 'presence-invisible'                              # Jabberd2
 NS_IQ             = 'iq'                                              # Jabberd2
 NS_JINGLE         ='urn:xmpp:jingle:1'                                # XEP-0166
@@ -92,7 +97,8 @@ NS_JINGLE_ERRORS  = 'urn:xmpp:jingle:errors:1'                        # XEP-0166
 NS_JINGLE_RTP     = 'urn:xmpp:jingle:apps:rtp:1'                      # XEP-0167
 NS_JINGLE_RTP_AUDIO = 'urn:xmpp:jingle:apps:rtp:audio'                # XEP-0167
 NS_JINGLE_RTP_VIDEO = 'urn:xmpp:jingle:apps:rtp:video'                # XEP-0167
-NS_JINGLE_FILE_TRANSFER ='urn:xmpp:jingle:apps:file-transfer:3'        # XEP-0234
+NS_JINGLE_FILE_TRANSFER = 'urn:xmpp:jingle:apps:file-transfer:3'      # XEP-0234
+NS_JINGLE_FILE_TRANSFER_5 = 'urn:xmpp:jingle:apps:file-transfer:5'    # XEP-0234
 NS_JINGLE_XTLS='urn:xmpp:jingle:security:xtls:0'                      # XTLS: EXPERIMENTAL security layer of jingle
 NS_JINGLE_RAW_UDP = 'urn:xmpp:jingle:transports:raw-udp:1'            # XEP-0177
 NS_JINGLE_ICE_UDP = 'urn:xmpp:jingle:transports:ice-udp:1'            # XEP-0176
@@ -101,6 +107,8 @@ NS_JINGLE_IBB     = 'urn:xmpp:jingle:transports:ibb:1'                # XEP-0261
 NS_LAST           = 'jabber:iq:last'
 NS_LOCATION       = 'http://jabber.org/protocol/geoloc'               # XEP-0080
 NS_MAM            = 'urn:xmpp:mam:0'                                  # XEP-0313
+NS_MAM_1          = 'urn:xmpp:mam:1'                                  # XEP-0313
+NS_MAM_2          = 'urn:xmpp:mam:2'                                  # XEP-0313
 NS_MESSAGE        = 'message'                                         # Jabberd2
 NS_MOOD           = 'http://jabber.org/protocol/mood'                 # XEP-0107
 NS_MUC            = 'http://jabber.org/protocol/muc'
@@ -111,6 +119,7 @@ NS_MUC_UNIQUE     = NS_MUC + '#unique'
 NS_MUC_CONFIG     = NS_MUC + '#roomconfig'
 NS_NICK           = 'http://jabber.org/protocol/nick'                 # XEP-0172
 NS_OFFLINE        = 'http://www.jabber.org/jeps/jep-0030.html'        # XEP-0013
+NS_OMEMO          = 'urn:xmpp:omemo:0'                                # XEP-0384
 NS_PHYSLOC        = 'http://jabber.org/protocol/physloc'              # XEP-0112
 NS_PING           = 'urn:xmpp:ping'                                   # XEP-0199
 NS_PRESENCE       = 'presence'                                        # Jabberd2
@@ -136,7 +145,9 @@ NS_SERVER         = 'jabber:server'
 NS_SESSION        = 'urn:ietf:params:xml:ns:xmpp-session'
 NS_SI             = 'http://jabber.org/protocol/si'                   # XEP-0096
 NS_SI_PUB         = 'http://jabber.org/protocol/sipub'                # XEP-0137
+NS_SID            = 'urn:xmpp:sid:0'                                  # XEP-0359
 NS_SIGNED         = 'jabber:x:signed'                                 # XEP-0027
+NS_SIMS           = 'urn:xmpp:sims:1'                                 # XEP-0385
 NS_SSN            = 'urn:xmpp:ssn'                                    # XEP-0155
 NS_STANZA_CRYPTO  = 'http://www.xmpp.org/extensions/xep-0200.html#ns' # XEP-0200
 NS_STANZAS        = 'urn:ietf:params:xml:ns:xmpp-stanzas'
@@ -164,12 +175,17 @@ NS_RECEIPTS       = 'urn:xmpp:receipts'
 NS_PUBKEY_PUBKEY  = 'urn:xmpp:pubkey:2'                                              # XEP-0189
 NS_PUBKEY_REVOKE  = 'urn:xmpp:revoke:2'
 NS_PUBKEY_ATTEST  = 'urn:xmpp:attest:2'
-NS_STREAM_MGMT    = 'urn:xmpp:sm:2'                                   # XEP-198
+NS_STREAM_MGMT    = 'urn:xmpp:sm:3'                                   # XEP-198
 NS_HASHES         = 'urn:xmpp:hashes:1'                               # XEP-300
-NS_HASHES_MD5     = 'urn:xmpp:hash-function-textual-names:md5'
-NS_HASHES_SHA1    = 'urn:xmpp:hash-function-textual-names:sha-1'
-NS_HASHES_SHA256  = 'urn:xmpp:hash-function-textual-names:sha-256'
-NS_HASHES_SHA512  = 'urn:xmpp:hash-function-textual-names:sha-512'
+NS_HASHES_2       = 'urn:xmpp:hashes:2'                               # XEP-300
+NS_HASHES_MD5     = 'urn:xmpp:hash-function-text-names:md5'
+NS_HASHES_SHA1    = 'urn:xmpp:hash-function-text-names:sha-1'
+NS_HASHES_SHA256  = 'urn:xmpp:hash-function-text-names:sha-256'
+NS_HASHES_SHA512  = 'urn:xmpp:hash-function-text-names:sha-512'
+NS_HASHES_SHA3_256 = 'urn:xmpp:hash-function-text-names:sha3-256'
+NS_HASHES_SHA3_512 = 'urn:xmpp:hash-function-text-names:sha3-512'
+NS_HASHES_BLAKE2B_256 = 'urn:xmpp:hash-function-text-names:id-blake2b256'
+NS_HASHES_BLAKE2B_512 = 'urn:xmpp:hash-function-text-names:id-blake2b512'
 
 #xmpp_stream_error_conditions = '''
 #bad-format --  --  -- The entity has sent XML that cannot be processed.
@@ -605,7 +621,8 @@ stream_exceptions = {'bad-format': BadFormat,
                     'unsupported-version': UnsupportedVersion,
                     'xml-not-well-formed': XMLNotWellFormed}
 
-class JID:
+
+class JID(object):
     """
     JID can be built from string, modified, compared, serialised into string
     """
@@ -726,7 +743,7 @@ class BOSHBody(Node):
     <body> tag that wraps usual XMPP stanzas in XMPP over BOSH
     """
 
-    def __init__(self, attrs={}, payload=[], node=None):
+    def __init__(self, attrs=None, payload=None, node=None):
         Node.__init__(self, tag='body', attrs=attrs, payload=payload, node=node)
         self.setNamespace(NS_HTTP_BIND)
 
@@ -737,8 +754,8 @@ class Protocol(Node):
     and messages
     """
 
-    def __init__(self, name=None, to=None, typ=None, frm=None, attrs={},
-                    payload=[], timestamp=None, xmlns=None, node=None):
+    def __init__(self, name=None, to=None, typ=None, frm=None, attrs=None,
+                    payload=None, timestamp=None, xmlns=None, node=None):
         """
         Constructor, name is the name of the stanza
         i.e. 'message' or 'presence'or 'iq'
@@ -819,6 +836,15 @@ class Protocol(Node):
             return self.timestamp
         return time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
 
+    def getJid(self):
+        """
+        Return the value of the 'jid' attribute
+        """
+        attr = self.getAttr('jid')
+        if attr:
+            return JID(attr)
+        return attr
+
     def getID(self):
         """
         Return the value of the 'id' attribute
@@ -931,6 +957,19 @@ class Protocol(Node):
                 props.append(prop)
         return props
 
+    def getTag(self, name, attrs=None, namespace=None, protocol=False):
+        """
+        Return the Node instance for the tag.
+        If protocol is True convert to a new Protocol/Message instance.
+        """
+        tag = Node.getTag(self, name, attrs, namespace)
+        if protocol and tag:
+            if name == 'message':
+                return Message(node=tag)
+            else:
+                return Protocol(node=tag)
+        return tag
+
     def __setitem__(self, item, val):
         """
         Set the item 'item' to the value 'val'
@@ -946,7 +985,7 @@ class Message(Protocol):
     """
 
     def __init__(self, to=None, body=None, xhtml=None, typ=None, subject=None,
-        attrs={}, frm=None, payload=[], timestamp=None, xmlns=NS_CLIENT,
+        attrs=None, frm=None, payload=None, timestamp=None, xmlns=NS_CLIENT,
         node=None):
         """
         You can specify recipient, text of message, type of message any
@@ -998,6 +1037,22 @@ class Message(Protocol):
         """
         return self.getTagData('thread')
 
+    def getOriginID(self):
+        """
+        Return origin-id of the message
+        """
+        return self.getTagAttr('origin-id', namespace=NS_SID, attr='id')
+
+    def getStanzaIDAttrs(self):
+        """
+        Return the stanza-id attributes of the message
+        """
+        try:
+            attrs = self.getTag('stanza-id', namespace=NS_SID).getAttrs()
+        except Exception:
+            return None, None
+        return attrs['id'], attrs['by']
+
     def setBody(self, val):
         """
         Set the text of the message"""
@@ -1035,6 +1090,12 @@ class Message(Protocol):
         """
         self.setTagData('thread', val)
 
+    def setOriginID(self, val):
+        """
+        Sets the origin-id of the message
+        """
+        self.setTag('origin-id', namespace=NS_SID, attrs={'id': val})
+
     def buildReply(self, text=None):
         """
         Builds and returns another message object with specified text. The to,
@@ -1061,7 +1122,7 @@ class Message(Protocol):
 class Presence(Protocol):
 
     def __init__(self, to=None, typ=None, priority=None, show=None, status=None,
-        attrs={}, frm=None, timestamp=None, payload=[], xmlns=NS_CLIENT,
+        attrs=None, frm=None, timestamp=None, payload=None, xmlns=NS_CLIENT,
         node=None):
         """
         You can specify recipient, type of message, priority, show and status
@@ -1182,8 +1243,8 @@ class Iq(Protocol):
     XMPP Iq object - get/set dialog mechanism
     """
 
-    def __init__(self, typ=None, queryNS=None, attrs={}, to=None, frm=None,
-                    payload=[], xmlns=NS_CLIENT, node=None):
+    def __init__(self, typ=None, queryNS=None, attrs=None, to=None, frm=None,
+                    payload=None, xmlns=NS_CLIENT, node=None):
         """
         You can specify type, query namespace any additional attributes,
         recipient of the iq, sender of the iq, any additional payload (f.e.
@@ -1339,6 +1400,62 @@ class Hashes(Node):
                 hash_ = hl.hexdigest()
         return hash_
 
+class Hashes2(Node):
+    """
+    Hash elements for various XEPs as defined in XEP-300
+    """
+
+    """
+    RECOMENDED HASH USE:
+    Algorithm     Support
+    MD2           MUST NOT
+    MD4           MUST NOT
+    MD5           MUST NOT
+    SHA-1         SHOULD NOT
+    SHA-256       MUST
+    SHA-512       SHOULD
+    SHA3-256      MUST
+    SHA3-512      SHOULD
+    BLAKE2b256    MUST
+    BLAKE2b512    SHOULD
+    """
+
+    supported = ('sha-256', 'sha-512', 'sha3-256', 'sha3-512', 'blake2b-256', 'blake2b-512')
+
+    def __init__(self, nsp=NS_HASHES):
+        Node.__init__(self, None, {}, [], None, None, False, None)
+        self.setNamespace(nsp)
+        self.setName('hash')
+
+    def calculateHash(self, algo, file_string):
+        """
+        Calculate the hash and add it. It is preferable doing it here
+        instead of doing it all over the place in Gajim.
+        """
+        hl = None
+        hash_ = None
+        if algo == 'sha-256':
+            hl = hashlib.sha256()
+        elif algo == 'sha-512':
+            hl = hashlib.sha512()
+        elif algo == 'sha3-256':
+            hl = hashlib.sha3_256()
+        elif algo == 'sha3-512':
+            hl = hashlib.sha3_512()
+        elif algo == 'blake2b-256':
+            hl = hashlib.blake2b(digest_size=32)
+        elif algo == 'blake2b-512':
+            hl = hashlib.blake2b(digest_size=64)
+        # file_string can be a string or a file
+        if hl is not None:
+            if isinstance(file_string, bytes):
+                hl.update(file_string)
+            else: # if it is a file
+                for line in file_string:
+                    hl.update(line)
+            hash_ = b64encode(hl.digest()).decode('ascii')
+        return hash_
+
     def addHash(self, hash_, algo):
         self.setAttr('algo', algo)
         self.setData(hash_)
@@ -1443,7 +1560,7 @@ class DataField(Node):
     """
 
     def __init__(self, name=None, value=None, typ=None, required=0, desc=None,
-                    options=[], node=None):
+                    options=None, node=None):
         """
         Create new data field of specified name,value and type
 
@@ -1592,7 +1709,7 @@ class DataForm(Node):
     Relevant XEPs: 0004, 0068, 0122. Can be used in disco, pub-sub and many
     other applications.
     """
-    def __init__(self, typ=None, data=[], title=None, node=None):
+    def __init__(self, typ=None, data=None, title=None, node=None):
         """
         Create new dataform of type 'typ'. 'data' is the list of DataField
         instances that this dataform contains, 'title' - the title string.  You
@@ -1620,18 +1737,19 @@ class DataForm(Node):
         self.setNamespace(NS_DATA)
         if title:
             self.setTitle(title)
-        if isinstance(data, dict):
-            newdata = []
-            for name in data.keys():
-                newdata.append(DataField(name, data[name]))
-            data = newdata
-        for child in data:
-            if child.__class__.__name__ == 'DataField':
-                self.kids.append(child)
-            elif isinstance(child, Node):
-                self.kids.append(DataField(node=child))
-            else: # Must be a string
-                self.addInstructions(child)
+        if data is not None:
+            if isinstance(data, dict):
+                newdata = []
+                for name in data.keys():
+                    newdata.append(DataField(name, data[name]))
+                data = newdata
+            for child in data:
+                if child.__class__.__name__ == 'DataField':
+                    self.kids.append(child)
+                elif isinstance(child, Node):
+                    self.kids.append(DataField(node=child))
+                else:  # Must be a string
+                    self.addInstructions(child)
 
     def getType(self):
         """
